/*
	Commands.cpp
	- Various Udp commands for Xicom UDP-based protocol.
*/


#include "XicomUdp.hpp"





/******************************************************************************
 ******************************************************************************
	SendUdpPacket
	- ip: ip address in string form such as "10.250.7.109".  Or NULL to broadcast.
	- buf: entire packet, ie starting with System-Type-Position and ending in \r\n.
	- len: # bytes in "buf" to send.  
 ******************************************************************************
*******************************************************************************/
void SendUdpPacket(sockaddr *ip, char *buf, int len)
  {
  sockaddr_in socket, *sin=&socket;
  int count, i;

  if (ip) sin = (struct sockaddr_in *) ip;
  else    sin->sin_addr.s_addr = INADDR_BROADCAST;
  sin->sin_family	= PF_INET;
  sin->sin_port 	= htons((short)XICOM_UDP_PORT);
	
  count = sendto(udp_socket, buf, len, 0, (struct sockaddr FAR *) sin, sizeof(sockaddr_in));
  if (count==SOCKET_ERROR) MessageBox(NULL, "Couldn't transmit packet", "Error", MB_OK);
  }




/************************************************************************************************************
 ************************************************************************************************************
  SendUdpCmd
  - Formats the given cmd into a Xicom-UDP-protocol packet, and sends it.
	- ip: target ip address, or NULL to broadcast the packet.
	  Typically one gets the IP address of the target unit upon receiving the unit's twice/sec status broadcast.
	- unit_type: 'T' = TWTA, 'E' = expansion controller, etc.
	- position: which position is the target for this command.
	  Amplifiers are 'A', 'B', C'.  Expansion controller is typically 'A'.
	- serno: the target serial number.  
	  This can be NULL, in which case the target will treat the field as a wild card, ignoring it.
	  However, it's good practise to use the target serial number:
	  	The serial# adds another layer of protection to ensure you are commanding the intended unit.
	  	Other parameters (ip address, system id/position) may not be unique in the event of a configuration error, 
	  	but the serial number is guaranteed to be unique to the unit.
	- cmd: command to be sent.  Example: "UA" = set automatic mode.
 ************************************************************************************************************
************************************************************************************************************/
void SendUdpCmd(sockaddr *ip, int unit_type, int position, char *serno, char *cmd)
  {
  char buf[256], *ptr=buf;
  int  i;

  if (strlen(cmd)>200) cmd[0]=0;	// bounds check!

  // Header
  *ptr++ = DigitToHex(system_id);	// Src system.
  *ptr++ = 'M';							// Src Type: M&C.
  *ptr++ = '1';							// Src Position (arbitrary).
  *ptr++ = '>';							// Packet Type: command.
  if (!serno || !strlen(serno)) 	// Dst serial number.
     *ptr++ = '0';						// wild card?
  else
     {
     strcpy(ptr, serno);
     ptr += strlen(serno);
     }
  *ptr++ = ';';							// terminate ser#.
  *ptr++ = DigitToHex(system_id); 	// Dst System: 99.9% of the time, this is the same as the src system.
  *ptr++ = unit_type; 					// Dst Typ: 'T'=Twta, 'K'=Klystron, 'E' Expansion controller, etc.
  *ptr++ = position;					// Dst Position: 'A', 'B', 'C', etc
  *ptr++ = ',';							// terminate header.

  // Payload.
  strcpy(ptr,cmd);	 
  ptr += strlen(cmd);

  // Terminate packet.
  *ptr++ = '\r';
  *ptr++ = '\n';
  *ptr = 0;

  // Send it.
  SendUdpPacket(ip, buf, ptr-buf);
  }
	
	



/******************************************************************************
 ******************************************************************************
	SetHpaInhibit
 ******************************************************************************
*******************************************************************************/
void SetHpaInhibit(int hpa, BOOL inhibit_on)
  {
	char cmd[20];
  s_amp_status *s=amp_state+hpa;
	if (!s->connected) 
	   {
		puts("\t... command not sent because unit is missing.");
		return;
		}
	if (inhibit_on) strcpy(cmd, "I1");
	else            strcpy(cmd, "I0");
	SendUdpCmd(&s->ip_address, s->tubetype, s->position, s->serial, cmd);
	}
	
	
/******************************************************************************
 ******************************************************************************
	SetExcWgSwitch
	- sw: 1, 2, etc.
	- pos: 'A' or 'B'.
 ******************************************************************************
*******************************************************************************/
void SetExcWgSwitch(int sw, int pos)
  {
	char cmd[20];
  s_exc_status *s=&exc_state;
	if (!s->connected) 
	   {
		puts("\t... command not sent because Expansion Controller is missing.");
		return;
		}
	sprintf(cmd, "B%d%c", sw, pos);
	SendUdpCmd(&s->ip_address, 'E', s->position, s->serial, cmd);
	}
			
