/*
	Parse.cpp
	- Parsing functions for Xicom UDP-based protocol.
*/


#include "XicomUdp.hpp"


// helper functions.
int  ExtractParameter(char *src, char *dst, int max, BOOL subparam=false);
void ParseSwitchStatus(BYTE &sw, BYTE ch, int bit_a, int bit_b);
void ParseVpcStatus(BYTE &vpc, BYTE ch);
BOOL Bit(BYTE val, int bit);
BOOL HexBit(BYTE val, int bit);
int  Hex2Number(char hex);







/******************************************************************************
 ******************************************************************************
	ParseUdpMsg
 ******************************************************************************
*******************************************************************************/
void ParseUdpMsg(char *msg, sockaddr *ip_addr)
	{
	int id = Hex2Number(msg[0]);		   // convert from ascii to number.
	if (id!=system_id) return;	// Not part of our system?
	
	if (msg[3]=='?') // A 2/sec status broadcast: 	 	STP?SSSSS,FFFFFFF,<Param ID#1><string>,<Param ID#n><string>
		switch (msg[1])
			{
			case 'E': 							// from an expansion controller.
				ParseExpCtlStatus(msg, ip_addr); 
				break;	
			case 'T': 							// from Twta, or...
			case 'K':							// from Kpa, or...
			case 'S':							// from SSPA
				ParseUdpStatusAmp(msg, ip_addr);
				break;	
			}

// 	if (msg[3]=='>') {} // A command, presumably not for us.
// 	if (msg[3]=='<') {} // A response, presumably not for us.
	}






/************************************************************************************************************
 ************************************************************************************************************
  ParseExpCtlStatus
  - Parses a UDP status broadcast from an EXPansion controller.
    message format:		1TC?SSSSS...[;]...,....
 ************************************************************************************************************
************************************************************************************************************/
void ParseExpCtlStatus(char *buf, sockaddr *ip_addr)
  {
	s_exc_status *ES=&exc_state;
  char *ptr=buf;
  char  tmp[80];
  int   index;

  ES->connected = POLL_TIMEOUT_COUNT;    // comm established.
	ES->position  = buf[2];				      // for sending cmds.
  ES->ip_address= *ip_addr;

  // extract serial#
	index=4;   // skip fixed id fields.
  index += ExtractParameter(ptr+index, ES->serial, 9);

  // extract and parse bitfields.
  ExtractParameter(ptr+index, tmp, 80);					         // byte 1
  ES->remotemode    	  = HexBit(ptr[index],0);                     
  ES->cmd_ack       	  = HexBit(ptr[index],1);
  ES->automode      	  = HexBit(ptr[index],2);
  ES->faults[A_CHECKSUM] = HexBit(ptr[index],3);
  ++index;																		// byte 2
  ES->faults[A_BATTERY]  = HexBit(ptr[index],0);
  ES->faults[A_VPCFAN]	  = HexBit(ptr[index],1);
  ++index;											                     // byte 3
  ES->faults[A_SUPPLY1]  = HexBit(ptr[index],0);
  ES->faults[A_SUPPLY2]  = HexBit(ptr[index],1);
  ES->faults[A_UPCONV1]  = HexBit(ptr[index],2);
  ES->faults[A_UPCONV2]  = HexBit(ptr[index],3);
  ++index;											                     // byte 4
	ES->upcon_enabled[0]   = HexBit(ptr[index],0);
	ES->upcon_enabled[1]   = HexBit(ptr[index],1);
	ES->ssa_inhibited[0]   = HexBit(ptr[index],2);
	ES->ssa_inhibited[1]   = HexBit(ptr[index],3);
  ++index;																		// byte 5
	ParseSwitchStatus(ES->insw[SWITCH1], ptr[index], 0, 1);
	ParseSwitchStatus(ES->insw[SWITCH2], ptr[index], 2, 3);
  ++index;																		// byte 6
	ParseSwitchStatus(ES->sw[SWITCH1], ptr[index], 0, 1);
	ParseSwitchStatus(ES->sw[SWITCH2], ptr[index], 2, 3);
  ++index;																		// byte 7
	ParseSwitchStatus(ES->sw[SWITCH3], ptr[index], 0, 1);
	ParseSwitchStatus(ES->sw[SWITCH4], ptr[index], 2, 3);
  ++index;																		// byte 8
	ParseSwitchStatus(ES->sw[SWITCH5], ptr[index], 0, 1);
	ParseSwitchStatus(ES->sw[SWITCH6], ptr[index], 2, 3);
  ++index;																		// byte 9
	ParseSwitchStatus(ES->sw[SWITCH7], ptr[index], 0, 1);
	ParseSwitchStatus(ES->sw[SWITCH8], ptr[index], 2, 3);
  ++index;																		// byte 10
	ParseSwitchStatus(ES->xfer[SWITCH1], ptr[index], 0, 1);
	ParseSwitchStatus(ES->xfer[SWITCH2], ptr[index], 2, 3);
  ++index;																		// byte 11
	ParseSwitchStatus(ES->xfer[SWITCH3], ptr[index], 0, 1);
	ParseSwitchStatus(ES->xfer[SWITCH4], ptr[index], 2, 3);
  ++index;																		// byte 12
	ES->sw_inh[SWITCH1] = HexBit(ptr[index],0);
	ES->sw_inh[SWITCH2] = HexBit(ptr[index],1);
	ES->sw_inh[SWITCH3] = HexBit(ptr[index],2);
	ES->sw_inh[SWITCH4] = HexBit(ptr[index],3);
  ++index;																		// byte 13
	ES->sw_inh[SWITCH5] = HexBit(ptr[index],0);
	ES->sw_inh[SWITCH6] = HexBit(ptr[index],1);
	ES->sw_inh[SWITCH7] = HexBit(ptr[index],2);
	ES->sw_inh[SWITCH8] = HexBit(ptr[index],3);
  ++index;																		// byte 14
	ParseVpcStatus(ES->vpc[SWITCH1], ptr[index]);
  ++index;																		// byte 15
	ParseVpcStatus(ES->vpc[SWITCH2], ptr[index]);
  ++index;																		// byte 16
	ParseVpcStatus(ES->vpc[SWITCH3], ptr[index]);
  //++index;	// future bits...?
  }





/************************************************************************************************************
 ************************************************************************************************************
  ParseUdpStatusAmp
  - Parses a UDP status broadcast from an HPA.
    1TC?SSSSS...[;]...,....
 ************************************************************************************************************
************************************************************************************************************/
void ParseUdpStatusAmp(char *buf, sockaddr *ip_addr)
  {
  s_amp_status *s;
  BOOL  overdrive;
  char *ptr=buf, *sptr; 
  int   len=strlen(ptr);
  char  tmp[80];
  char *tp=tmp+1;  // point to start of parameter following parameter id.
  int   index, count, amp;
	
  switch (buf[2])	// which amp is this?
     {
     case 'A': case '1': amp=AMP_A; break;
     case 'B': case '2': amp=AMP_B; break;
     case 'C': case '3': amp=AMP_C; break;
     default: return;	// out of bounds.
     }
	s = amp_state+amp;

  s->connected = POLL_TIMEOUT_COUNT;     // comm established with amp.
  s->tubetype  = buf[1];
  s->position  = buf[2];
  s->ip_address= *ip_addr;

  // extract serial#
  index=4;   // skip fixed id fields.
  index += ExtractParameter(ptr+index, s->serial, 9);

  // extract and parse bitfields.
  count = ExtractParameter(ptr+index, tmp, 80);
  s->remotemode    = HexBit(ptr[index],0);                     // byte 1
  s->hv_on         = HexBit(ptr[index],1);
  s->ftd           = HexBit(ptr[index],2);
  s->rf_inhibited  = HexBit(ptr[index],3);
  ++index;
  s->cmd_ack       = HexBit(ptr[index],0);                     // byte 2
  s->cpwr_en       = HexBit(ptr[index],1);
  s->hv_selected   = HexBit(ptr[index],2);
  s->htr_standby   = HexBit(ptr[index],3);
  ++index;
  s->upc_enabled   = HexBit(ptr[index],0);                     // byte 3
  s->upc_suspended = HexBit(ptr[index],1);
  s->faults[A_BUC_OVERDRIVE] = HexBit(ptr[index],2);
  s->automode      = HexBit(ptr[index],3);
  ++index;
  int s1a          = HexBit(ptr[index],0);                     // byte 4
  int s1b          = HexBit(ptr[index],1);
  int s2a          = HexBit(ptr[index],2);
  int s2b          = HexBit(ptr[index],3);
  if (s1a!=s1b) s->sw1 = s1a?'A': 'B';	// Note: A 1:2 amp-B can't see this switch.
  else          s->sw1 = 0;
  if (s2a!=s2b) s->sw2 = s2a?'A': 'B';	// Note: A 1:2 amp-A can't see this switch.
  else          s->sw2 = 0;
  ++index;
  s->inh_selected  = HexBit(ptr[index],0);              // byte 5
  s->ext_inhibit   = HexBit(ptr[index],1);
  s->sw1_inhibit   = HexBit(ptr[index],2);
  s->sw2_inhibit   = HexBit(ptr[index],3);
  ++index;
  s->odrive_inhibit= HexBit(ptr[index],0);              // byte 6
  s->wg_inhibit    = HexBit(ptr[index],1);
  s->sbuc_warmup   = HexBit(ptr[index],2);
  // bit3: undefined.
  ++index;
  s->sumfault                 = HexBit(ptr[index],0);   // byte 7
  s->faults[F_LOWLINE        ] = HexBit(ptr[index],1);
  s->faults[F_OVERVOLT       ] = HexBit(ptr[index],2);
  s->faults[F_UNDERVOLT      ] = HexBit(ptr[index],3);
  ++index;
  s->faults[F_OVERTEMP       ] = HexBit(ptr[index],0);   // byte 8
  s->faults[F_MAXPWR         ] = HexBit(ptr[index],1);
  s->faults[F_MINPWR         ] = HexBit(ptr[index],2);
  s->faults[F_RFLPWR         ] = HexBit(ptr[index],3);
  ++index;
  s->faults[F_FAST_VSWR      ] = HexBit(ptr[index],0);   // byte 9
  s->faults[F_SW_OVERDRIVE   ] = HexBit(ptr[index],1);
  s->faults[F_HELIX_ARC      ] = HexBit(ptr[index],2);
  s->faults[F_HELIX_ARC_L    ] = HexBit(ptr[index],3);
  ++index;
  s->faults[F_WG_ARC         ] = HexBit(ptr[index],0);   // byte 10
  s->faults[F_UPCONVERTER    ] = HexBit(ptr[index],1);
  s->faults[F_IN_OVERDRIVE   ] = HexBit(ptr[index],2);
  s->faults[F_IN_OVERDRIVE_L ] = HexBit(ptr[index],3);
  ++index;
  s->faults[F_EXT_INTERLOCK  ] = HexBit(ptr[index],0);   // byte 11
  s->faults[F_TWT_TH_ILOCK   ] = HexBit(ptr[index],1);
  s->faults[F_COVER_INTERLOCK] = HexBit(ptr[index],2);
  s->faults[F_HEATER         ] = HexBit(ptr[index],3);
  ++index;
  s->faults[F_24V            ] = HexBit(ptr[index],0);    // byte 12
  s->faults[F_EXT_REF        ] = HexBit(ptr[index],1);
  s->faults[F_EXT_REF_LOCK   ] = HexBit(ptr[index],2);
  s->faults[F_LINEARIZER     ] = HexBit(ptr[index],3);
  ++index;
  s->faults[F_AUX_SUPPLY     ] = HexBit(ptr[index],0);    // byte 13
  s->faults[F_WG_PRESSURE    ] = HexBit(ptr[index],1);
  ++index;
  s->sumalarm                  = HexBit(ptr[index],0);     // byte 14
  s->faults[A_LOW_DCV        ] = HexBit(ptr[index],1);
  s->faults[A_MAXPWR         ] = HexBit(ptr[index],2);
  s->faults[A_MINPWR         ] = HexBit(ptr[index],3);
  ++index;
  s->faults[A_RFLPWR         ] = HexBit(ptr[index],0);    // byte 15
  s->faults[A_REDUCE_DRIVE   ] = HexBit(ptr[index],1);
  s->faults[A_RAISE_DRIVE    ] = HexBit(ptr[index],2);
  overdrive                    = HexBit(ptr[index],3);
  ++index;
  s->cpwr_suspend              = HexBit(ptr[index],0);    // byte 16
  s->faults[A_TEMPERATURE    ] = HexBit(ptr[index],1);
  s->faults[A_EXT_REF        ] = HexBit(ptr[index],2);
  s->flog_changed              = HexBit(ptr[index],3);
  ++index;                                         		// Byte 17
	overdrive |= HexBit(ptr[index],3);	// v31: "tube overdrive" alarm, consolidated with byte15/bit3.
  s->faults[A_OVERDRIVE   	] = overdrive;
  ++index;                                          		// Byte 18
  ++index;                                         		// Byte 19
  s->faults[F_PWR_SUPPLY	   ] = HexBit(ptr[index],0);
  ++index;                                          		// Byte 20
  s->faults[F_PWR_SUPPLY2    ] = HexBit(ptr[index],0);
  s->faults[F_SSPA_5V	 	   ] = HexBit(ptr[index],1);
  s->faults[F_SSPA_5V2	      ] = HexBit(ptr[index],2);
  s->faults[F_SSPA_7V	 	   ] = HexBit(ptr[index],3);
  ++index;                                          		// Byte 21
  s->faults[F_SSPA_7V2	 	   ] = HexBit(ptr[index],0);
  s->faults[F_SSPA_10V	 	   ] = HexBit(ptr[index],1);
  s->faults[F_SSPA_10V2	 	] = HexBit(ptr[index],2);
  s->faults[F_CABLE_INTERLOCK2]= HexBit(ptr[index],3);
  ++index;

  // skip any future bits.
  if (count>21) index += count-21;

  // Extract and parse remaining fields which arrive in arbitrary order, each starting with a unique parameter id.
  s->cathode_current  = 0;
  s->enable_cathode_i = false;  // unless we find it below.
  s->has_atten        = false;   // unless we find it below.
  s->sub_band			  = 0;
  s->batt_mv			  = 0xFFFF;	// aka parameter not supported.
  while (index<len-2)
     {
     index += ExtractParameter(ptr+index, tmp, 80);
     switch (tmp[0])
        {
        case '+': s->batt_mv          = atoi(tp); 		break;
        case 'A': s->has_atten        = true;
                  s->atten            = atof(tp);       break;
        case 'G': s->cpwr_targ        = atof(tp);       break;
        case 'C': s->cathode_current  = atoi(tp);
                  s->enable_cathode_i = true;           break;
        case 'I': s->helix_i          = atof(tp);       break;
        case 'J': s->heater_v         = atof(tp);       break;
        case 'M': s->h_hours          = atoi(tp);       break;
        case 'N': s->b_hours          = atoi(tp);       break;
        case 'P': s->fpower           = atof(tp);
           sptr = strchr(tp, '<');
           if (sptr) s->power_floor   = atof(sptr+1); // Below readable?  Aka forward power floor.
           break; 
        case 'Q': s->rpower           = atof(tp); 
           sptr = strchr(tp, '<');
           if (sptr) s->rvs_floor     = atof(sptr+1); // below readable? Aka reverse power floor.
           break; 
        case 's': s->sub_band 			= atoi(tp);			break;
        case 'T': s->temp             = atoi(tp);       break;
        case 'V': s->helix_v          = atof(tp);       break;
        case 'H': 
           switch (*tp)
              {
              case 'A': s->usr_max_trip = EVENT_ALARM;   break;
              case 'F': s->usr_max_trip = EVENT_FAULT;   break;
              case 'D': s->usr_max_trip = EVENT_DEFAULT; break;
              }
           s->usr_max_limit = atof(tp+1);
           break;
        case 'L':
           switch (*tp)
              {
              case 'A': s->usr_min_trip = EVENT_ALARM;   break;
              case 'F': s->usr_min_trip = EVENT_FAULT;   break;
              case 'D': s->usr_min_trip = EVENT_DEFAULT; break;
              }
           s->usr_min_limit = atof(tp+1);
           break;
        case 'R':
           switch (*tp)
              {
              case 'A': s->usr_rfl_trip = EVENT_ALARM;   break;
              case 'F': s->usr_rfl_trip = EVENT_FAULT;   break;
              case 'D': s->usr_rfl_trip = EVENT_DEFAULT; break;
              }
           s->usr_rfl_limit = atof(tp+1);
           break;
        }
     }

  // Inferred states.
  s->cpwr_suspend |= s->cpwr_en && (!s->hv_on || s->rf_inhibited);
  if (s->sbuc_warmup) s->hv_on = false;   // hv is technically 'on' but unusable without buc lock during smart buc warmup.
  }












/************************************************************************************************************
 ************************************************************************************************************
  ParseSwitchStatus
  - Helper function, extracts switch position from UDP status bitfields.
	- sw: alias to a switch variable to receive the parsed status 'A'/'B'/0.
	- ch: the status byte to be decoded.
	- bit_a, bit_b: the respective locations (0-3) of the position-A and position-B bits.  
 ************************************************************************************************************
************************************************************************************************************/
void ParseSwitchStatus(BYTE &sw, BYTE ch, int bit_a, int bit_b)
	{
	int posA = HexBit(ch, bit_a); 
	int posB = HexBit(ch, bit_b);
  if (posA!=posB) sw = posA? 'A': 'B';
  else        	 sw = 0;
	}



/************************************************************************************************************
 ************************************************************************************************************
  ParseVpcStatus
  - Helper function, extracts vpc position from UDP status bitfields.
	- vpc: alias to a variable to receive the parsed status 'A'/'B'/'C'/'D'/0.
	- ch: the status byte to be decoded.  Assumes the entire byte is used to code the status.
 ************************************************************************************************************
************************************************************************************************************/
void ParseVpcStatus(BYTE &vpc, BYTE ch)
	{
	if 	  (HexBit(ch,0)) vpc = 'A';
	else if (HexBit(ch,1)) vpc = 'B';
	else if (HexBit(ch,2)) vpc = 'C';
	else if (HexBit(ch,3)) vpc = 'D';
	else						  vpc =  0 ;
	}




/************************************************************************************************************
 ************************************************************************************************************
  Bit
	- returns TRUE if the specified bit is set.
 ************************************************************************************************************
************************************************************************************************************/
BOOL Bit(BYTE  val,  int  bit) { return ((val & (1<<bit))? TRUE: FALSE); }


/************************************************************************************************************
 ************************************************************************************************************
  HexBit
	- returns TRUE if the specified bit is set in the given hexbyte "val".
    Examples: Hexbit('1',0) returns true because hexbyte '1'=0001, where bit0=true.
              Hexbit('C',3) returns true because hexbyte 'C'=1100, where the bit3=true.
  - val: must be one of '0'->'9', or 'A'-'F'.
  - bit: must be 0-3.
 ************************************************************************************************************
************************************************************************************************************/
BOOL HexBit(BYTE val, int bit) 
  {
  if (val>='0' && val<='9') return Bit(val-'0',bit);
  if (val>='A' && val<='F') return Bit(val-55 ,bit); // convert to range 10-15, given 'A'=65.
  return false;
  }


/************************************************************************************************************
 ************************************************************************************************************
  ExtractParameter
  - For parsing parameters from Xicom UDP-based protocol.
    Retrieves a delimited parameter from buffer "src", storing it in dst.
    Characters are copied from the source string until reaching ',' [cr], [lf], or the end of the string.
    Whitespace is ignored.
    The resulting string is null-terminated.
  - src: buffer containing the parameter.
  - dst: buffer to store the extracted parameter.
  - max: size of "dst" buffer.
  - returns the number of characters read from src regardless of whether they were stored in dst.  
    The returned count includes the terminating ','.
  - subparam: if TRUE, semicolons (';') also count as a terminator.  Default is FALSE.
 ************************************************************************************************************
************************************************************************************************************/
int ExtractParameter(char *src, char *dst, int max, BOOL subparam)
  {
  BOOL copy=true;
  int count;
  if (--max<0) return 0;  // leave room for null terminator.
  for (count=0; *src; ++src)
     {
     ++count;
     if (*src==' ' || *src=='\t') continue; // ignore whitespace.
     if (*src==',' || *src=='\r' || *src=='\n' || count>max) copy=false;   // stop copying?
     if (subparam && *src==';') copy=false;
     if (copy) *dst++ = *src;
     if (*src==',' || *src=='\r' || *src=='\n') break;
     if (subparam && *src==';') break;
     }
  *dst=0;
  return count;
  }


/************************************************************************************************************
 ************************************************************************************************************
  Hex2Number
  - Converts a hex character '0-9', 'A'-'F', to a number (0x00-0x0F).
 ************************************************************************************************************
************************************************************************************************************/
int Hex2Number(char hex)
  {
  if (hex>='0' && hex<='9') return hex-'0';
  if (hex>='a' && hex<='f') return hex-'a'+10;
  if (hex>='A' && hex<='F') return hex-'A'+10;
  return 0;
  }


