/*
	Windows command-line program to illustrate Xicom UDP protocol.
*/


#include "XicomUdp.hpp"



SOCKET  udp_socket;
char udp_buf[1024];	// buffer for processing received packets.
int  system_id=1;	// Can be 1-15.  Matches user-selected system id for the controller/HPAs.

s_exc_status exc_state;					// expansion controller status.
s_amp_status amp_state[TOTAL_AMPS];	// array of amp structures.


void ShowMenu();
BOOL ProcessKey(int ch);
void ShowHpaStatus(int hpa);
void ShowExcStatus();





/******************************************************************************
 ******************************************************************************
	main
 ******************************************************************************
*******************************************************************************/
void main()
	{
	InitStatus();
	InitUdp();
	puts("************** UMON ... Xicom Udp-based protocol demo **************");
	ShowMenu();
	
	while (1) 
		{
		CheckUdp();
		TrackConnectivity();
		Sleep(500); // establishes interval for TrackConnectivity().
		if (kbhit() && ProcessKey(getch()))
		   break;
		}
	
	CloseUdp();
	}





/******************************************************************************
 ******************************************************************************
	InitStatus
 ******************************************************************************
*******************************************************************************/
void InitStatus()
	{
	memset(&exc_state, 0, sizeof(exc_state));
	memset( amp_state, 0, sizeof(amp_state));
	}



/******************************************************************************
 ******************************************************************************
	InitUdp
	- Enable reception of Xicom UDP packets.
	  Uses native Windows calls.  (Most modern compilers offer an easier way.)
 ******************************************************************************
*******************************************************************************/
void InitUdp()
	{
  sockaddr_in udp_socketinfo;

  // Init windows sockets.
  WSADATA wsaData;
  WSAStartup(MAKEWORD(1,1), &wsaData);

  // Create a "socket" for UDP, disable blocking, allow broadcasting.
  unsigned long nonblocking=1;
  BOOL allow_broadcast=true, reuse_socket=true;
  udp_socket = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
  ioctlsocket(udp_socket, FIONBIO, &nonblocking);
  setsockopt(udp_socket, SOL_SOCKET, SO_BROADCAST, (char *)&allow_broadcast, sizeof(BOOL));
  setsockopt(udp_socket, SOL_SOCKET, SO_REUSEADDR, (char *)&reuse_socket,    sizeof(BOOL));

  // Start receiving packets on XICOM_UDP_PORT.
  udp_socketinfo.sin_family		= PF_INET;
  udp_socketinfo.sin_addr.s_addr = INADDR_ANY;
  udp_socketinfo.sin_port 		   = htons((short)XICOM_UDP_PORT);
  if (bind(udp_socket, (LPSOCKADDR)&udp_socketinfo, sizeof(udp_socketinfo))==SOCKET_ERROR)
     MessageBox(NULL, "Couldn't open UDP.", "Error", MB_OK);
	}



/******************************************************************************
 ******************************************************************************
	CloseUdp
 ******************************************************************************
 ******************************************************************************/
void CloseUdp()
	{
	closesocket(udp_socket);
	udp_socket = INVALID_SOCKET;
	WSACleanup();
	}




/******************************************************************************
 ******************************************************************************
	CheckUdp
	- Checks for inbound Xicom UDP packets.
 ******************************************************************************
 ******************************************************************************/
void CheckUdp()
	{
	struct sockaddr ip_addr;
	int  addr_size = sizeof(struct sockaddr);
	int  count;

  // read packets until no more are available.
	while ((count = recvfrom(udp_socket,udp_buf,1023,0,&ip_addr,&addr_size)) > 0)
		{
		if (count<5) continue;			// Too short to be a valid message?  Ignore.
		udp_buf[count]=0;					// ensure null termination.
		ParseUdpMsg(udp_buf, &ip_addr);
		}
	}




/******************************************************************************
 ******************************************************************************
	TrackConnectivity
	- call this 2/sec to track which devices are responding.
	- Assumes that parse routines are setting device "connected" upon seeing a msg.
 ******************************************************************************
*******************************************************************************/
void TrackConnectivity()
  {
  // exp ctl
  s_exc_status *es=&exc_state;
  if (--es->connected<0) es->connected=0;
  if (es->connected && !es->was_connected)                       // just found exc?
     printf("Found Expansion Controller, serial#='%s'.\r\n", es->serial);
  if (es->was_connected && !es->connected)                       // just lost exc?
     {
     puts("Lost status from Expansion Controller.");
     memset(es, 0, sizeof(s_exc_status));                        // reset all stats.
     }
  es->was_connected = es->connected;

  // amps
  s_amp_status *as;
  for (int hpa=AMP_A; hpa<TOTAL_AMPS; hpa++)
     {
     as = amp_state+hpa;
     if (--as->connected<0) as->connected=0;
     if (as->connected && !as->was_connected)                       // just found hpa?
        printf("Found Hpa-%c, serial#='%s'.\r\n", hpa+'A', as->serial);
     if (as->was_connected && !as->connected)                       // just lost hpa?
        {
        printf("Lost status from Hpa-%c.\r\n", hpa+'A');
        memset(as, 0, sizeof(s_amp_status));                        // reset all stats.
        }
     as->was_connected = as->connected;
     }
	}





/******************************************************************************
 ******************************************************************************
	ShowMenu
 ******************************************************************************
*******************************************************************************/
void ShowMenu()
	{
	puts("\tMENU");
	printf("\tS - Set system ID (%d)\r\n", system_id);
	puts("\tA - Hpa-A status \t1 - Inhibit Hpa-A RF \t4 - Enable Hpa-A RF");
	puts("\tB - Hpa-B status \t2 - Inhibit Hpa-B RF \t5 - Enable Hpa-B RF");
	puts("\tC - Hpa-C status \t3 - Inhibit Hpa-C RF \t6 - Enable Hpa-C RF");
	puts("\tX - ExpCtl status \t7 - Sw1 pos A \t\t8 - Sw1 pos B");
	puts("\tQ - Quit\r\n");
  }



/******************************************************************************
 ******************************************************************************
	ProcessKey
	- returns true if the program should exit.
 ******************************************************************************
*******************************************************************************/
BOOL ProcessKey(int ch) 
  {
	int val;
	char buf[256];
	
	switch (ch)
	   {
		case 'S': case 's':
			printf("Enter new system ID (1-15):");
			gets(buf);
			val = atoi(buf);
			if (val>15) 
			   {
				puts("A valid System ID is between 1 and 15");
				break;
				}
			if (val && val!=system_id) InitStatus();
			if (val) system_id=val;
			printf("Using system id=%d\r\n", system_id);
		   break;
			
		case 'A': case 'B': case 'C': 
		   ShowHpaStatus(ch-'A');
		   break;
	   case 'a': case 'b': case 'c':
		   ShowHpaStatus(ch-'a');
		   break;
     case '1': case '2': case '3':
		   printf("Sending RF Inhibit to amp-%c\r\n", ch-'1'+'A');
		   SetHpaInhibit(ch-'1', true);
			break;
		case '4': case '5': case '6':
		   printf("Sending RF Enable to amp-%c\r\n", ch-'4'+'A');
		   SetHpaInhibit(ch-'4', false);
			break;
     
		case '7':
		   puts("Sending command 'Switch1->PositionA' to Expansion Controller");
		   SetExcWgSwitch(1, 'A');
			break;
		case '8':
		   puts("Sending command 'Switch1->PositionB' to Expansion Controller");
		   SetExcWgSwitch(1, 'B');
			break;
			
     case 'X': case 'x':
		   ShowExcStatus();
			break;
		
		case 'Q': case 'q':
		   return true;  // user wants to exit.
		}
	
	return false;  // not time to terminate program.
	}
	
	
	
	







/******************************************************************************
 ******************************************************************************
	ShowExcStatus
 ******************************************************************************
*******************************************************************************/
void ShowExcStatus()
  {
  struct s_exc_status *s = &exc_state;
	char str[40];
	
	if (!s->connected)
	   {
		puts("Expansion Controller...NOT FOUND");
		return;
		}

	puts("\r\n********************* Expansion Controller Status ********************");
	
	printf(" Serial#=%s ", s->serial);
  sockaddr_in *ip = (struct sockaddr_in *) &s->ip_address;
	printf("\t\t IP addr=%s", inet_ntoa(ip->sin_addr));
	puts("");

	if (s->remotemode)       printf(" REMOTE mode");
	else                     printf(" LOCAL mode ");
	if (s->automode)         printf("\t\t Autoswitch: ON ");
	else                     printf("\t\t Autoswitch: OFF");
	puts("");
	
	if (!s->sw[SWITCH1]) printf(" Sw1=? ");
	else                 printf(" Sw1=%c ", s->sw[SWITCH1]);
	if (!s->sw[SWITCH2]) printf("\t\t\t Sw2=? ");
	else                 printf("\t\t\t Sw2=%c ", s->sw[SWITCH2]);
	puts("");
	
	if (s->faults[A_SUPPLY1]) printf(" Supply1 FAULT ");
	else                      printf(" Supply1 OK    ");
	if (s->faults[A_SUPPLY2]) printf("\t\t Supply2 FAULT");
	else                      printf("\t\t Supply2 OK   ");
	puts("");

	puts("**********************************************************************\r\n");
	ShowMenu();
	}





/******************************************************************************
 ******************************************************************************
	ShowHpaStatus
	- Shows various status of typical interest.
 ******************************************************************************
*******************************************************************************/
void ShowHpaStatus(int hpa)
  {
  struct s_amp_status *s = amp_state+hpa;
	char str[40];
	
	if (!s->connected)
	   {
		printf("Hpa-%c...NOT FOUND\r\n", hpa+'A');
		return;
		}

	switch (s->tubetype)
	   {
		case 'T': strcpy(str, "TWTA"); break;
		case 'K': strcpy(str, "KPA"); break;
		case 'S': strcpy(str, "SSPA"); break;
		default : strcpy(str, "UNIT"); break;
		}
	printf("\r\n**************************** %s-%c Status ****************************\r\n", str, s->position);
	// cout << "HPA-" << s->position << " Status";
	
	printf(" Serial#=%s ", s->serial);
  sockaddr_in *ip = (struct sockaddr_in *) &s->ip_address;
	printf("\t\t IP addr=%s", inet_ntoa(ip->sin_addr));
	puts("");

  if      (s->ftd)         printf(" FTD Warmup ");
	else if (s->hv_on)       printf(" HV is ON   ");
	else if (s->hv_selected) printf(" HV pending ");
	else                     printf(" Standby    ");
	if (s->rf_inhibited)     printf("\t\t RF Inhibited ");
	else                     printf("\t\t RF Enabled   ");
	puts("");

	if (s->remotemode)       printf(" REMOTE mode");
	else                     printf(" LOCAL mode ");
	if (s->automode)         printf("\t\t Autoswitch: ON ");
	else                     printf("\t\t Autoswitch: OFF");
	puts("");

  printf(" Fwd Pwr: %.1f dBm \t Rev Pwr: %.1f dBm \t ", s->fpower, s->rpower);
	if (s->has_atten) printf("Attenuation: %.1f dB", s->atten);
	puts("");

	printf(" Temp: %.1f C \t\t Helix V: %.1f kV \t Helix I: %.1f", s->temp, s->helix_v, s->helix_i);   // or beamV/beamI if kpa.
	puts("");
	
  printf(" Heater hours=%d \t Beam hours=%d", s->h_hours, s->b_hours);
	puts("");

	if (s->sumfault)         printf(" SUMMARY FAULT ");
	else                     printf(" No Faults     ");
	if (s->sumfault)         printf("\t\t Summary Alarm ");
	else                     printf("\t\t No Warnings   ");
	puts("");
	
	puts("**********************************************************************\r\n");
	ShowMenu();
	}
